/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#include <el/ext.h>

#include "patch-bfi_int.h"

#include "elf.h"

struct Section {
	String Name;
	Buf Mb;
};


void PatchBfiInt::Patch(const Buf& mb) {
	if (mb.Size < 1024)
		Throw(E_FAIL);
	Elf32_Ehdr& hdr = *(Elf32_Ehdr*)mb.P;
	if (memcmp(hdr.e_ident, "\x7F\x45\x4C\x46\x01\x01\x01\x64", 8))
		Throw(E_FAIL);
	if (0 == hdr.e_shoff)
		return;
	Elf32_Shdr& sech = *(Elf32_Shdr*)(mb.P+hdr.e_shoff+hdr.e_shstrndx*hdr.e_shentsize);
	const byte *sh = mb.P + hdr.e_shoff;

	vector<Section> vSec;
	for (int i=0; i<hdr.e_shnum; ++i) {
		Elf32_Shdr& entry = *(Elf32_Shdr*)(sh+i*hdr.e_shentsize);
		Section sec;
		sec.Name = (const char*)mb.P+sech.sh_offset+entry.sh_name;
		sec.Mb = Buf(mb.P+entry.sh_offset, entry.sh_size);
		vSec.push_back(sec);
	}
	int n = 0;
	Section *sec = 0;
	for (int i=0; i<vSec.size(); ++i) {
		if (vSec[i].Name == ".text") {
			if (++n == 2)
				sec = &vSec[i];
		}
	}
	if (n != 2)
		return;
	int nPatch = 0;
	for (int i=0; i<sec->Mb.Size/8; ++i) {
		UInt64& instr = *(UInt64*)(sec->Mb.P+i*8);
		if ((instr & 0x9003f00002001000ULL) == 0x0001a00000000000ULL) {
			nPatch += 1;
            instr ^=  (0x0001a00000000000ULL ^ 0x0000c00000000000ULL); // BFI_INT
		}
	}
//	if (nPatch < 60)
//		Throw(E_FAIL);
}


