/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#include <el/ext.h>

#include "num.h"


using namespace Ext;

namespace Ext { namespace Num {

Bn::Bn(const BigInteger& bn) {
	int n = (bn.Length+8)/8;
	byte *p = (byte*)alloca(n);
	bn.ToBytes(p, n);
	std::reverse(p, p+n);
	SslCheck(m_bn = ::BN_bin2bn(p, n, 0));
}

BigInteger Bn::ToBigInteger() const {
	int n = BN_num_bytes(m_bn);
	byte *p = (byte*)alloca(n+1);
	if (::BN_bn2bin(m_bn, p) != n)
		Throw(E_FAIL);
	std::reverse(p, p+n);
	p[n] = 0;
	return BigInteger(p, n+1);
}

Bn Bn::FromBinary(const ConstBuf& cbuf) {
	BIGNUM *a = ::BN_bin2bn(cbuf.P, cbuf.Size, 0);
	SslCheck(a);
	return Bn(a);
}

void Bn::ToBinary(byte *p, size_t n) const {
	int a = BN_num_bytes(m_bn);
	if (a > n)
		Throw(E_INVALIDARG);
	memset(p, 0, n-a);
	SslCheck(::BN_bn2bin(m_bn, p+(n-a)));
}


}} // Ext::Num::


