/*######     Copyright (c) 1997-2013 Ufasoft  http://ufasoft.com  mailto:support@ufasoft.com,  Sergey Pavlov  mailto:dev@ufasoft.com #######################################
#                                                                                                                                                                          #
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation;  #
# either version 3, or (at your option) any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the      #
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should have received a copy of the GNU #
# General Public License along with this program; If not, see <http://www.gnu.org/licenses/>                                                                               #
##########################################################################################################################################################################*/

#pragma once

#include <el/bignum.h>
#include <el/crypto/ext-openssl.h>

#include <openssl/bn.h>

// Modular arithmetic


namespace Ext { namespace Num {

using namespace Ext::Crypto;

class BnCtx {
public:
	BnCtx() {
		SslCheck(m_ctx = ::BN_CTX_new());
	}

	~BnCtx() {
		::BN_CTX_free(m_ctx);
	}

	operator BN_CTX*() {
		return m_ctx;
	}
private:
	BN_CTX *m_ctx;

	BnCtx(const BnCtx& bnCtx);
	BnCtx& operator=(const BnCtx& bnCtx);
};

class Bn {
public:
	Bn() {
		SslCheck(m_bn = ::BN_new());
	}

	Bn(const Bn& bn) {
		SslCheck(::BN_dup(bn.m_bn));
	}

	explicit Bn(const BigInteger& bn);

	~Bn() {
		::BN_clear_free(m_bn);
	}

	Bn& operator=(const Bn& bn) {
		::BN_clear_free(m_bn);
		SslCheck(::BN_dup(bn.m_bn));
		return *this;
	}

	operator const BIGNUM *() const {
		return m_bn;
	}

	BigInteger ToBigInteger() const;
	static Bn FromBinary(const ConstBuf& cbuf);
	void ToBinary(byte *p, size_t n) const;
	BIGNUM *Ref() { return m_bn; }	
private:
	BIGNUM *m_bn;

	explicit Bn(BIGNUM *bn)
		:	m_bn(bn)
	{}
};

}} // Ext::Num::

